
/*

OWL Lite and OWL DL

Abstract Syntax, Mapping Rules, Concrete Syntax

Author: Carroll
Based on: AS&amp;S editor's draft 24 January 2003 Patel-Schneider
With significant modification.

*/

:-op(1200,xfx,'::=').
:- dynamic (::=)/2.
:- discontiguous (::=)/2.

terminal(uriref).
terminal(lexicalForm).
terminal(language).
terminal(0).
terminal(1).


/* Notation:
   ::= rewrite arrow
   {}   zero or more
   [] zero or one
   + terminal symbol
   (o,m,dt, ...)   IDs
*/
ontology ::=
   ontology( { directive } ).

comment(rationale,
   ['When dealing with imports, the resulting ontology may many ',link(header),' blocks, with many
     different ',link(o),' URI references. Hence, the change from the published AS&amp;S.' ]).

directive ::=
   header .

directive ::=
   axiom .

directive ::=
   fact .

header ::=
   header( o, {metaPropValue}, {annotation} ).


metaPropValue ::= 
   metaPropValue( m, o ).
annotation ::=
   annotation( sdp, dataLiteral ).
comment(note,
   ['Wherever in an abstract syntax document the following rule is used,
     the ',next(individual),' is also treated as a top-level ',link(fact),'.']).

annotation ::=
   annotation( sop, individual ).

fact ::= 
   individual .

fact ::= allDifferentIndividuals( {i} ).


individual ::=
   individual( [i],  { type( description ) }, 
   { propertyValue }, {individualComparison} ).


propertyValue ::=
    propertyValue( sop, individual ).
propertyValue ::=
    propertyValue( cop, individual ).
propertyValue ::=
    propertyValue( tp, individual ).
propertyValue ::=
   propertyValue( sdp, dataLiteral ).
propertyValue ::=
   propertyValue( cdp, dataLiteral ).

%type ::= c.
%type ::= restriction.

description ::= c.
description ::= restriction.

comment(note,
   [ 'This maps directly onto RDF literals and has the same constraints and definitions.',
     'See [RDF Concepts]. Note, specifically that there the lexicalForm is in Unicode Normal 
      Form C, and that the language identifier is an RFC 3066 language tag normalized to lower
      case.'
   ]).


dataLiteral ::=
      dataLiteral(lexicalForm,[language],[dt]).

individualComparison ::=
   sameIndividual(individual).
individualComparison ::=
   differentIndividuals(individual).

comment(note,
   [ next({classComparison}), ' is in this rule, primarily for use in OWL DL. ',
    'In the OWL Lite context, it is clearer to have an EquivalentClasses axiom (as in ',
    'the published abstract syntax.']).

axiom ::=
   class(c, modality, {annotation},
                    [+deprecated], {super(description)},
     { classComparison }
   ).
modality ::= +partial.
modality ::= +complete.

classComparison ::= sameClassAs(c).

comment(rationale, ['This differs from the published AS&amp;S in that each ',link(restriction),' may only
                      contain one condition, not an arbitrary number. The expressive power is unchanged,
                      since every use may be repeated arbitrarily. This makes it easier to express the
                      mapping to RDF in which each condition becomes a separate blank node.']). 

restriction ::= restriction( cdp, allValuesFrom( dataRange ), {restrictionComparison} ).
restriction ::= restriction( cdp, someValuesFrom( dataRange ), {restrictionComparison} ).
restriction ::= restriction( cdp, cardinality(smallInt), {restrictionComparison} ).
restriction ::= restriction( cdp, minCardinality(smallInt), {restrictionComparison} ).
restriction ::= restriction( cdp, maxCardinality(smallInt), {restrictionComparison} ).

restriction ::= restriction( cop, allValuesFrom( classInRestriction ), {restrictionComparison} ).
restriction ::= restriction( cop, someValuesFrom( classInRestriction  ), {restrictionComparison} ).
restriction ::= restriction( cop, cardinality(smallInt), {restrictionComparison} ).
restriction ::= restriction( cop, minCardinality(smallInt), {restrictionComparison} ).
restriction ::= restriction( cop, maxCardinality(smallInt), {restrictionComparison} ).

restriction ::= restriction( tp, allValuesFrom( classInRestriction  ), {restrictionComparison} ).
restriction ::= restriction( tp, someValuesFrom( classInRestriction  ), {restrictionComparison} ).

axiom ::= restriction .
comment(rationale, 
  ['The extra level here is used for an extension in OWL DL.']).
classInRestriction ::= c .

restrictionComparison ::= +none .

smallInt ::= 0.
smallInt ::= 1.

dataRange ::= dt.
dataRange ::= dr.
axiom ::=
  datatypeProperty( sdp, { annotation },
                    [+deprecated],
                    { super(sdp) } ).

comment(rationale,
  [ 'For ease of writing the document, I have made the property axioms
     identical in OWL Lite and OWL DL, but kept the same semantics by
     having the additional rules for dataRange and classInRestriction 
     in OWL DL' ]).
axiom ::=
  datatypeProperty( cdp, { annotation },
                    [+deprecated],
                    { super(sdp) },
                    { super(cdp) },
                    { domain(classInRestriction) },
                    { range(dataRange) }, [+functional] ).

axiom ::= individualProperty( sop, { annotation },
                    [+deprecated],
                    { super(sop) } ).
axiom ::=
  individualProperty( cop, { annotation },
                    [+deprecated],
                    { super(sop) },
                    { super(cop) },
                    { super(tp) },
                    { domain(classInRestriction) },
                    { range(classInRestriction) }, 
                    { inverse(cop) },
                    [+functional], 
                    [+inverseFunctional], 
                    [+symmetric] ).
axiom ::=
  individualProperty( tp, { annotation },
                    [+deprecated],
                    { super(sop) },
                    { super(tp) },
                    { domain(classInRestriction) },
                    { range(classInRestriction) }, 
                    { inverse(tp) }, 
                    [+transitive] ).

axiom ::=
   equivalentProperties(sop,sop,{sop}).
axiom ::=
   equivalentProperties(tp,tp,{tp}).
axiom ::=
   equivalentProperties(cop,cop,{cop}).
axiom ::=
   equivalentProperties(sdp,sdp,{sdp}).
axiom ::=
   equivalentProperties(cdp,cdp,{cdp}).

axiom ::=
   datatypeDeclaration( dt ).

axiom ::= description.