

/* 
Rule format:
 one of:
AbstractSyntaxExpr -> Triples.
AbstractSyntaxExpr -> Triples; MainNode .
SubjNode+AbstractSyntax -> Triples.


Internal syntax
   id-subscript
   note(link)/thingy
*/

:-op(150,xfy,@).
:-op(150,xfy,^^).

comment(1, heading(1,'URI References')).
comment(2, 'The following rules are ones that use a node provided from higher
 in the abstract syntax tree. This node is shown as <code><strong>@</strong></code>.
It is always the subject of the triples.').

o(O) -> 
        t(O,rdf:type,owl:ontology);      O.
dt(DT) -> 
         t(DT,rdf:type,rdfs:datatype);  DT.

dt(rdf:xMLLiteral) -> true;              rdf:xMLLiteral.

dt(xsd:Builtin) -> true;                 xsd:Builtin.

%dt(Any) ->
%         t(Any, rdf:type, rdfs:datatype ); Any.


c(owl:thing) -> true;                    owl:thing.
c(owl:nothing) -> true;                    owl:nothing.

c(C) -> 
         t(C,rdf:type,owl:class),
         +[t(C,rdf:type,rdfs:class)];     C.

i(I) -> true;                            I.

m(owl:Builtin) -> true;                  owl:Builtin.
m(rdfs:Builtin) -> true;                 rdfs:Builtin .
dr(rdfs:literal) -> true;                rdfs:literal.
dr(DR) -> 
         note(dataRange)/t(DR,rdf:type,owl:dataRange);   DR.

sdp(DP) -> 
          t(DP,rdf:type,owl:datatypeProperty),
          +[t(DP,rdf:type,rdf:property)];
                                          DP.

sdp(rdfs:Builtin) -> true;    rdfs:Builtin .

cdp(DP) -> 
          t(DP,rdf:type,owl:datatypeProperty),
          +[t(DP,rdf:type,rdf:property)]; DP.

sop(OP) -> 
          t(OP,rdf:type,owl:objectProperty),
          +[t(OP,rdf:type,rdf:property)];
                                          OP.

sop(rdfs:Builtin) -> true; rdfs:Builtin .

cop(OP) -> 
          note(prop)/(+[t(OP,rdf:type,owl:objectProperty)]),
          +[t(OP,rdf:type,rdf:property)]; OP.

tp(OP) -> 
          note(prop)/(+[t(OP,rdf:type,owl:objectProperty)]),
          +[t(OP,rdf:type,rdf:property)]; OP.
comment(1, heading(1,'Literals')).
dataLiteral(lexicalForm,language,dt)->true; lexicalForm@language^^x(dt).
dataLiteral(lexicalForm,dt)->true; lexicalForm^^x(dt).
dataLiteral(lexicalForm,language)->true; lexicalForm@language.
dataLiteral(lexicalForm)->true; lexicalForm.
         

comment(1, heading(1,'Ontologies')).

ontology( { directive } ) -> 
               {note(directive)/x(directive)}.


comment(2, heading(1,'Header and Annotations')).
header(o,{metaPropValue},{annotation}) -> 
          t(x(o),rdf:type,owl:ontology),
          {x(x(o),metaPropValue)},
          {x(x(o),annotation)}.

D+metaPropValue( m, o ) -> 
          t(D,x(m),x(o) ).

D + annotation( sdp, dataLiteral ) -> 
          t(D,x(sdp),x(dataLiteral)).
D + annotation( sop, individual ) ->
          t(D,x(sop),x(individual) ).

comment(1, heading(1,individuals)).
comment(1, heading(2,'Named Individuals')).
comment(1, 'The interaction between these two individual rules make it obligatory 
to declare at least one type for a named individual.
The default is <code>owl:thing</code>').

individual( i, type( description-1 ), {type( description-2 )}, 
            {propertyValue}, {individualComparison} ) ->
       t(x(i),rdf:type,x(description-1)),
       {t(x(i),rdf:type,x(description-2))},
       {x(x(i),propertyValue)},
       {x(x(i),individualComparison)};            x(i).


individual( i, {type( description )}, 
            {propertyValue}, {individualComparison} ) ->
       t(x(i),rdf:type,owl:thing),
       {t(x(i),rdf:type,x(description))},
       {x(x(i),propertyValue)},
       {x(x(i),individualComparison)};            x(i).

comment(1, heading(2,'Unnamed Individuals')).

 
individual(   type( description-1 ), {type( description-2 )}, 
            {propertyValue}, {individualComparison} ) ->
       t(blank,rdf:type,x(description-1)),
       {t(blank,rdf:type,x(description-2))},
       {x(blank,propertyValue)},
       {x(blank,individualComparison)};            blank.


individual(  {type( description )}, 
            {propertyValue}, {individualComparison} ) ->
       t(blank,rdf:type,owl:thing),
       {t(blank,rdf:type,x(description))},
       {x(blank,propertyValue)},
       {x(blank,individualComparison)};            blank.   

comment(2, heading(1,'Property Values')).

D+propertyValue(sop,individual) ->
       t(D,x(sop),x(individual)).
D+propertyValue(cop,individual) ->
       t(D,x(cop),x(individual)).
D+propertyValue(tp,individual) ->
       t(D,x(tp),x(individual)).
D+propertyValue(sdp,dataLiteral) ->
       t(D,x(sdp),x(dataLiteral)).
D+propertyValue(cdp,dataLiteral) ->
       t(D,x(cdp),x(dataLiteral)).


comment(2, heading(1,'Comparison of Individuals')).

D+sameIndividual(individual)->
   t(D,owl:sameIndividualAs,x(individual)).
D+differentIndividuals(individual)->
   t(D,owl:differentFrom,x(individual)).

comment(1, heading(1,'Classes')).


class(c, +partial, {annotation},
                    [+deprecated], {super(description)},
     { classComparison }
   ) ->
   t(x(c),rdf:type,owl:class),
   +[t(x(c),rdf:type,rdfs:class)],
   {x(x(c),annotation)},
   [t(x(c),rdf:type,owl:deprecatedClass)],
   {t(x(c),rdfs:subClassOf,x(description))},
   {x(x(c),classComparison)};     x(c).

class(c, +complete, {annotation},
                    [+deprecated], {super(description)},
     { classComparison }
   ) ->
   t(x(c),rdf:type,owl:class),
   +[t(x(c),rdf:type,rdfs:class)],
   {x(x(c),annotation)},
   [t(x(c),rdf:type,owl:deprecatedClass)],
   t(x(c),owl:intersectionOf,x(seq({description}))),
   {x(x(c),classComparison)};     x(c).

class(c, +complete, {annotation}, 
                    [+deprecated], super(unionOf({description}, {classComparison-2})),
     { classComparison-1 }
   ) ->
   t(x(c),rdf:type,owl:class),
   +[t(x(c),rdf:type,rdfs:class)],
   {x(x(c),annotation)},
   [t(x(c),rdf:type,owl:deprecatedClass)],
   t(x(c),owl:unionOf,x(seq({description}))),
   {x(x(c),classComparison-2)},
   {x(x(c),classComparison-1)};     x(c).


class(c, +complete, {annotation},
                    [+deprecated],  super(complementOf(description, {classComparison-2})),
     { classComparison-1 }
   ) ->
   t(x(c),rdf:type,owl:class),
   +[t(x(c),rdf:type,rdfs:class)],
   {x(x(c),annotation)},
   [t(x(c),rdf:type,owl:deprecatedClass)],
   t(x(c),owl:complementOf,x(description)),
   {x(x(c),classComparison-2)},
   {x(x(c),classComparison-1)};     x(c).
      
comment(1, heading(2,'Enumerated Classes')).
enumeratedClass(c, { annotation }, 
                    [+deprecated], {i}, { classComparison } )->
   t(x(c),rdf:type,owl:class),
   +[t(x(c),rdf:type,rdfs:class)],
   {x(x(c),annotation)},
   [t(x(c),rdf:type,owl:deprecatedClass)],
   t(x(c),owl:oneOf,x(seq({i}))),
   {x(x(c),classComparison)};     x(c).


comment(2, heading(1,'Comparison of Classes')).

D+sameClassAs(c) ->
   t(D,owl:sameClassAs,x(c)).
D+sameClassAs(description)->
   t(D,owl:sameClassAs,x(description)).
D+disjointWith(description)->
   t(D,owl:disjointWith,x(description)).
D+subClassOf(description)->
   t(D,rdfs:subClassOf,x(description)).
_+ ( +none) -> true.

comment(1, heading(1,'Restrctions')).

comment(1, heading(2,'Restrictions on Datatype Properties')).


restriction( cdp, allValuesFrom( dataRange ), {restrictionComparison} ) ->
   t(blank,rdf:type,owl:restriction),
   t(blank,owl:onProperty,x(cdp)),
   t(blank,owl:allValuesFrom,x(dataRange)),
   {x(blank,restrictionComparison)};  
                                        blank.

restriction( cdp, someValuesFrom( dataRange ), {restrictionComparison} ) ->
   t(blank,rdf:type,owl:restriction),
   t(blank,owl:onProperty,x(cdp)),
   t(blank,owl:someValuesFrom,x(dataRange)),
   {x(blank,restrictionComparison)};  
                                        blank.    

restriction( cdp, cardinality(smallInt), {restrictionComparison} ) ->
   t(blank,rdf:type,owl:restriction),
   t(blank,owl:onProperty,x(cdp)),
   t(blank,owl:cardinality,x(smallInt)),
   {x(blank,restrictionComparison)};  
                                        blank.    
restriction( cdp, minCardinality(smallInt), {restrictionComparison} ) ->
   t(blank,rdf:type,owl:restriction),
   t(blank,owl:onProperty,x(cdp)),
   t(blank,owl:minCardinality,x(smallInt)),
   {x(blank,restrictionComparison)};  
                                        blank.     
restriction( cdp, maxCardinality(smallInt), {restrictionComparison} ) ->
   t(blank,rdf:type,owl:restriction),
   t(blank,owl:onProperty,x(cdp)),
   t(blank,owl:maxCardinality,x(smallInt)),
   {x(blank,restrictionComparison)};  
                                        blank.     
restriction( cdp, value(dataLiteral), {restrictionComparison} ) ->
   t(blank,rdf:type,owl:restriction),
   t(blank,owl:onProperty,x(cdp)),
   t(blank,owl:hasValue,x(dataLiteral)),
   {x(blank,restrictionComparison)};  
                                        blank.  


comment(1, heading(2,'Restrictions on Object Properties')).

restriction( cop, allValuesFrom( classInRestriction ), {restrictionComparison} ) ->
   t(blank,rdf:type,owl:restriction),
   t(blank,owl:onProperty,x(cop)),
   t(blank,owl:allValuesFrom,x(classInRestriction)),
   {x(blank,restrictionComparison)};  
                                        blank.

restriction( cop, someValuesFrom( classInRestriction ), {restrictionComparison} ) ->
   t(blank,rdf:type,owl:restriction),
   t(blank,owl:onProperty,x(cop)),
   t(blank,owl:someValuesFrom,x(classInRestriction)),
   {x(blank,restrictionComparison)};  
                                        blank.    

restriction( cop, cardinality(smallInt), {restrictionComparison} ) ->
   t(blank,rdf:type,owl:restriction),
   t(blank,owl:onProperty,x(cop)),
   t(blank,owl:cardinality,x(smallInt)),
   {x(blank,restrictionComparison)};  
                                        blank.    
restriction( cop, minCardinality(smallInt), {restrictionComparison} ) ->
   t(blank,rdf:type,owl:restriction),
   t(blank,owl:onProperty,x(cop)),
   t(blank,owl:minCardinality,x(smallInt)),
   {x(blank,restrictionComparison)};  
                                        blank.     
restriction( cop, maxCardinality(smallInt), {restrictionComparison} ) ->
   t(blank,rdf:type,owl:restriction),
   t(blank,owl:onProperty,x(cop)),
   t(blank,owl:maxCardinality,x(smallInt)),
   {x(blank,restrictionComparison)};  
                                        blank.      
restriction( cop, value(i), {restrictionComparison} ) ->
   t(blank,rdf:type,owl:restriction),
   t(blank,owl:onProperty,x(cop)),
   t(blank,owl:hasValue,x(i)),
   {x(blank,restrictionComparison)};  
                                        blank.  

comment(1, heading(2,'Restrictions on Transitive Properties')).


restriction( tp, allValuesFrom( classInRestriction ), {restrictionComparison} ) ->
   t(blank,rdf:type,owl:restriction),
   t(blank,owl:onProperty,x(tp)),
   t(blank,owl:allValuesFrom,x(classInRestriction)),
   {x(blank,restrictionComparison)};  
                                        blank.

restriction( tp, someValuesFrom( classInRestriction ), {restrictionComparison} ) ->
   t(blank,rdf:type,owl:restriction),
   t(blank,owl:onProperty,x(tp)),
   t(blank,owl:someValuesFrom,x(classInRestriction)),
   {x(blank,restrictionComparison)};  
                                        blank.     
restriction( tp, value(i), {restrictionComparison} ) ->
   t(blank,rdf:type,owl:restriction),
   t(blank,owl:onProperty,x(tp)),
   t(blank,owl:hasValue,x(i)),
   {x(blank,restrictionComparison)};  
                                        blank.  

comment(1, heading(2,'Numbers in Cardinalities')).
0 -> true; 0^^xsd:nonNegativeInteger.
1 -> true; 1^^xsd:nonNegativeInteger.
naturalNumber -> true; naturalNumber^^xsd:nonNegativeInteger.


comment(1, heading(1,'Property Axioms' )).
datatypeProperty( sdp-1, { annotation },[+deprecated],
                    { super(sdp-2) } )  ->
   t(x(sdp-1),rdf:type,owl:datatypeProperty),
  +[t(x(sdp-1),rdf:type,rdf:property)],
   { x(x(sdp-1),annotation) },
   [t(x(sdp-1),rdf:type,owl:deprecatedProperty)],
   { t(x(sdp-1),rdfs:subPropertyOf,x(sdp-2)) }.
  
datatypeProperty( cdp-1, { annotation },[+deprecated],
                    { super(sdp) },
                    { super(cdp-2) },
                    { domain(classInRestriction) },
                    { range(dataRange) }, [+functional] ) ->
   t(x(cdp-1),rdf:type,owl:datatypeProperty),
  +[t(x(cdp-1),rdf:type,rdf:property)],
   { x(x(cdp-1),annotation) },
   [t(x(cdp-1),rdf:type,owl:deprecatedProperty)],
   { t(x(cdp-1),rdfs:subPropertyOf,x(sdp)) },
   { t(x(cdp-1),rdfs:subPropertyOf,x(cdp-2)) },
   { t(x(cdp-1), rdfs:domain, x(classInRestriction)) },
   { t(x(cdp-1), rdfs:range, x(dataRange) ) },
   [ t(x(cdp-1), rdf:type, owl:functionalProperty) ].

individualProperty( sop-1, { annotation },[+deprecated],
                    { super(sop-2) } ) ->
   t(x(sop-1),rdf:type,owl:objectProperty),
  +[t(x(sop-1),rdf:type,rdf:property)],
   { x(x(sop-1),annotation) },
   [t(x(sop-1),rdf:type,owl:deprecatedProperty)],
   { t(x(sop-1),rdfs:subPropertyOf,x(sop-2)) }.


individualProperty( cop-1, { annotation },[+deprecated],
                    { super(sop) },
                    { super(cop-2) },
                    { super(tp) },
                    { domain(classInRestriction-1) },
                    { range(classInRestriction-2) }, 
                    { inverse(cop-3) },
                    [+functional], 
                    [+inverseFunctional], 
                    [+symmetric]) ->
   note(prop)/(+[t(x(cop-1),rdf:type,owl:objectProperty)]),
  +[t(x(cop-1),rdf:type,rdf:property)],
   { x(x(cop-1),annotation) },
   [t(x(cop-1),rdf:type,owl:deprecatedProperty)],
   { t(x(cop-1),rdfs:subPropertyOf,x(sop)) },
   { t(x(cop-1),rdfs:subPropertyOf,x(cop-2)) },
   { t(x(cop-1),rdfs:subPropertyOf,x(tp)) },
   { t(x(cop-1), rdfs:domain, x(classInRestriction-1)) },
   { t(x(cop-1), rdfs:range, x(classInRestriction-2)) },
   { t(x(cop-1), owl:inverseOf, x(cop-3)) },
   [t(x(cop-1),rdf:type,owl:functionalProperty)],
   [t(x(cop-1),rdf:type,owl:inverseFunctionalProperty)],
   [t(x(cop-1),rdf:type,owl:symmetricProperty)].

individualProperty( tp-1, { annotation },[+deprecated],
                    { super(sop) },
                    { super(tp-2) },
                    { domain(classInRestriction-1) },
                    { range(classInRestriction-2) }, 
                    { inverse(tp-3) },
                    [+transitive]) ->
   note(prop)/( +[t(x(tp-1),rdf:type,owl:objectProperty)]),
  +[t(x(tp-1),rdf:type,rdf:property)],
   { x(x(tp-1),annotation) },
   [t(x(tp-1),rdf:type,owl:deprecatedProperty)],
   { t(x(tp-1),rdfs:subPropertyOf,x(sop)) },
   { t(x(tp-1),rdfs:subPropertyOf,x(tp-2)) },
   { t(x(tp-1), rdfs:domain, x(classInRestriction-1)) },
   { t(x(tp-1), rdfs:range, x(classInRestriction-2)) },
   { t(x(tp-1), owl:inverseOf, x(tp-3)) },
   [t(x(tp-1),rdf:type,owl:transitiveProperty)].

allDifferentIndividuals( {i} ) ->
   t(blank,rdf:type,owl:allDifferent),
   t(blank,owl:distinctMembers,x(seq({i}))); blank .

comment(1, heading(1,'Equivalent Properties')).
equivalentProperties(sop-1,sop-2,{sop-3}) ->
   t(x(sop-1),owl:samePropertyAs,x(sop-2)),
   {t(x(sop-1),owl:samePropertyAs,x(sop-3))}.

equivalentProperties(tp-1,tp-2,{tp-3}) ->
   t(x(tp-1),owl:samePropertyAs,x(tp-2)),
   {t(x(tp-1),owl:samePropertyAs,x(tp-3))}.

equivalentProperties(cop-1,cop-2,{cop-3}) ->
   t(x(cop-1),owl:samePropertyAs,x(cop-2)),
   {t(x(cop-1),owl:samePropertyAs,x(cop-3))}.


equivalentProperties(sdp-1,sdp-2,{sdp-3}) ->
   t(x(sdp-1),owl:samePropertyAs,x(sdp-2)),
   {t(x(sdp-1),owl:samePropertyAs,x(sdp-3))}.

equivalentProperties(cdp-1,cdp-2,{cdp-3}) ->
   t(x(cdp-1),owl:samePropertyAs,x(cdp-2)),
   {t(x(cdp-1),owl:samePropertyAs,x(cdp-3))}.

%datatypeDeclaration(dt) -> x(dt).
comment(1, heading(1,'Descriptions')).



intersectionOf({description}, {classComparison}) ->
   t(blank,rdf:type,owl:class),
   t(blank,owl:intersectionOf,x(seq({description}))),
   {x(blank,classComparison)};     blank.

unionOf({description}, {classComparison}) ->
   t(blank,rdf:type,owl:class),
   t(blank,owl:unionOf,x(seq({description}))),
   {x(blank,classComparison)};     blank.

complementOf(description, {classComparison}) ->
   t(blank,rdf:type,owl:class),
   t(blank,owl:complementOf,x(description)),
   {x(blank,classComparison)};     blank.

oneOf({i}, {classComparison}) ->
   t(blank,rdf:type,owl:class),
   t(blank,owl:oneOf,x(seq({i}))),
   {x(blank,classComparison)};     blank.

comment(1, heading(1,'DataRanges')).

oneOf( {dataLiteral} ) ->
   t(blank,rdf:type,owl:dataRange),
   t(blank,owl:oneOf,x(seq({dataLiteral})));     blank.
 
dataRange( dr, {annotation}, {dataLiteral} ) ->
   t(x(dr),rdf:type,owl:dataRange),
   {x(x(dr),annotation)},
   t(x(dr),owl:oneOf,x(seq({dataLiteral})));     x(dr).
dataRange( dr, +declaration, {annotation} ) ->
   t(x(dr),rdf:type,owl:dataRange),
   {x(x(dr),annotation)};     x(dr).
   

seq({[]}) -> true; rdf:nil .
seq(description-1,{description-2}) -> 
    t(blank,rdf:type,rdf:list),
    t(blank,rdf:first,x(description-1)),
    t(blank,rdf:rest,x(seq({description-2}))); blank.

seq(dataLiteral-1,{dataLiteral-2}) -> 
    t(blank,rdf:type,rdf:list),
    t(blank,rdf:first,x(dataLiteral-1)),
    t(blank,rdf:rest,x(seq({dataLiteral-2}))); blank.

seq(i-1,{i-2}) -> 
    t(blank,rdf:type,rdf:list),
    t(blank,rdf:first,x(i-1)),
    t(blank,rdf:rest,x(seq({i-2}))); blank.

/*
seq(type-1,{type-2}) -> 
    t(blank,rdf:type,rdf:list),
    t(blank,rdf:first,x(type-1)),
    t(blank,rdf:rest,x(seq({type-2}))); blank.
*/

